@@ -47,28 +47,29 @@
  *
  */
 
-#include "miniserver.h"
+	#include "miniserver.h"
 
-#include "httpreadwrite.h"
-#include "ithread.h"
-#include "ssdplib.h"
-#include "statcodes.h"
-#include "ThreadPool.h"
-#include "unixutil.h" /* for socklen_t, EAFNOSUPPORT */
-#include "upnpapi.h"
-#include "upnputil.h"
+	#include "ThreadPool.h"
+	#include "httpreadwrite.h"
+	#include "ithread.h"
+	#include "ssdplib.h"
+	#include "statcodes.h"
+	#include "unixutil.h" /* for socklen_t, EAFNOSUPPORT */
+	#include "upnpapi.h"
+	#include "upnputil.h"
 
-#include <assert.h>
-#include <errno.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <sys/types.h>
+	#include <assert.h>
+	#include <errno.h>
+	#include <stdio.h>
+	#include <stdlib.h>
+	#include <string.h>
+	#include <sys/types.h>
 
-/*! . */
-#define APPLICATION_LISTENING_PORT 49152
+	/*! . */
+	#define APPLICATION_LISTENING_PORT 49152
 
-struct mserv_request_t {
+struct mserv_request_t
+{
 	/*! Connection handle. */
 	SOCKET connfd;
 	/*! . */
@@ -76,7 +77,8 @@ struct mserv_request_t {
 };
 
 /*! . */
-typedef enum {
+typedef enum
+{
 	/*! . */
 	MSERV_IDLE,
 	/*! . */
@@ -92,26 +94,132 @@ uint16_t miniStopSockPort;
  * module vars
  */
 static MiniServerState gMServState = MSERV_IDLE;
-#ifdef INTERNAL_WEB_SERVER
+	#ifdef INTERNAL_WEB_SERVER
 static MiniServerCallback gGetCallback = NULL;
 static MiniServerCallback gSoapCallback = NULL;
 static MiniServerCallback gGenaCallback = NULL;
 
+static const int ENABLE_IPV6 =
+		#ifdef UPNP_ENABLE_IPV6
+	1;
+		#else
+	0;
+		#endif
+
+static const int MINISERVER_REUSEADDR =
+		#ifdef UPNP_MINISERVER_REUSEADDR
+	1;
+		#else
+	0;
+		#endif
+
+struct s_SocketStuff
+{
+	int ip_version;
+	const char *text_addr;
+	struct sockaddr_storage ss;
+	union
+	{
+		struct sockaddr *serverAddr;
+		struct sockaddr_in *serverAddr4;
+		struct sockaddr_in6 *serverAddr6;
+	};
+	SOCKET fd;
+	uint16_t try_port;
+	uint16_t actual_port;
+	socklen_t address_len;
+};
+
 void SetHTTPGetCallback(MiniServerCallback callback)
 {
 	gGetCallback = callback;
 }
 
-#ifdef INCLUDE_DEVICE_APIS
-void SetSoapCallback(MiniServerCallback callback)
-{
-	gSoapCallback = callback;
-}
-#endif /* INCLUDE_DEVICE_APIS */
+		#ifdef INCLUDE_DEVICE_APIS
+void SetSoapCallback(MiniServerCallback callback) { gSoapCallback = callback; }
+		#endif /* INCLUDE_DEVICE_APIS */
 
-void SetGenaCallback(MiniServerCallback callback)
+void SetGenaCallback(MiniServerCallback callback) { gGenaCallback = callback; }
+
+static int host_header_is_numeric(char *host_port, size_t host_port_size)
 {
-	gGenaCallback = callback;
+	int rc = 0;
+	struct in6_addr addr;
+	char *s;
+
+	/* Remove the port part. */
+	s = host_port + host_port_size - 1;
+	while (s != host_port && *s != ']' && *s != ':') {
+		--s;
+	}
+	if (*s == ':') {
+		*s = 0;
+	} else {
+		s = host_port + host_port_size;
+	}
+
+	/* Try IPV4 */
+	rc = inet_pton(AF_INET, host_port, &addr);
+	if (rc == 1) {
+		goto ExitFunction;
+	}
+	/* Try IPV6 */
+	/* Check for and remove the square brackets. */
+	if (strlen(host_port) < 3 || host_port[0] != '[' || *(s - 1) != ']') {
+		rc = 0;
+		goto ExitFunction;
+	}
+	*(s - 1) = '\0';
+	rc = inet_pton(AF_INET6, host_port + 1, &addr) == 1;
+
+ExitFunction:
+	return rc;
+}
+
+static int getNumericHostRedirection(
+	int socket, char *host_port, size_t hp_size)
+{
+	int rc;
+	const char *rcp;
+	struct sockaddr_storage addr;
+	struct sockaddr_in *addr4 = (struct sockaddr_in *)&addr;
+	struct sockaddr_in6 *addr6 = (struct sockaddr_in6 *)&addr;
+	socklen_t addr_len = sizeof addr;
+	in_port_t port;
+	char host[NAME_SIZE];
+	int n;
+
+	rc = getsockname(socket, (struct sockaddr *)&addr, &addr_len);
+	if (rc) {
+		goto ExitFunction;
+	}
+	switch (addr.ss_family) {
+	case AF_INET6:
+		rcp = inet_ntop(AF_INET6, &addr6->sin6_addr, host, sizeof host);
+		if (!rcp) {
+			rc = -1;
+			goto ExitFunction;
+		}
+		port = ntohs(addr6->sin6_port);
+		n = snprintf(host_port, hp_size, "[%s]:%d", host, port);
+		break;
+	case AF_INET:
+	default:
+		rcp = inet_ntop(AF_INET, &addr4->sin_addr, host, sizeof host);
+		if (!rcp) {
+			rc = -1;
+			goto ExitFunction;
+		}
+		port = ntohs(addr4->sin_port);
+		n = snprintf(host_port, hp_size, "%s:%d", host, port);
+		break;
+	}
+	if (n < 0) {
+		rc = -1;
+	}
+
+ExitFunction:
+	return rc == 0;
 }
 
 /*!
@@ -125,21 +233,39 @@ static int dispatch_request(
 	/*! [in] HTTP parser object. */
 	http_parser_t *hparser)
 {
+	memptr header;
+	size_t min_size;
+	http_message_t *request;
 	MiniServerCallback callback;
+	WebCallback_HostValidate host_validate_callback = 0;
+	void *cookie = 0;
+	int rc = UPNP_E_SUCCESS;
+	/* If it does not fit in here, it is likely invalid anyway. */
+	char host_port[NAME_SIZE];
 
 	switch (hparser->msg.method) {
 	/* Soap Call */
 	case SOAPMETHOD_POST:
 	case HTTPMETHOD_MPOST:
 		callback = gSoapCallback;
+		UpnpPrintf(UPNP_INFO,
+			MSERV,
+			__FILE__,
+			__LINE__,
+			"miniserver %d: got SOAP msg\n",
+			info->socket);
 		break;
 	/* Gena Call */
 	case HTTPMETHOD_NOTIFY:
 	case HTTPMETHOD_SUBSCRIBE:
 	case HTTPMETHOD_UNSUBSCRIBE:
-		UpnpPrintf(UPNP_INFO, MSERV, __FILE__, __LINE__,
-			"miniserver %d: got GENA msg\n", info->socket);
-			callback = gGenaCallback;
+		callback = gGenaCallback;
+		UpnpPrintf(UPNP_INFO,
+			MSERV,
+			__FILE__,
+			__LINE__,
+			"miniserver %d: got GENA msg\n",
+			info->socket);
 		break;
 	/* HTTP server call */
 	case HTTPMETHOD_GET:
@@ -147,16 +273,82 @@ static int dispatch_request(
 	case HTTPMETHOD_HEAD:
 	case HTTPMETHOD_SIMPLEGET:
 		callback = gGetCallback;
+		host_validate_callback = gWebCallback_HostValidate;
+		cookie = gWebCallback_HostValidateCookie;
+		UpnpPrintf(UPNP_INFO,
+			MSERV,
+			__FILE__,
+			__LINE__,
+			"miniserver %d: got WEB server msg\n",
+			info->socket);
 		break;
 	default:
-		callback = NULL;
+		callback = 0;
 	}
-	if (callback == NULL) {
-		return HTTP_INTERNAL_SERVER_ERROR;
+	if (!callback) {
+		rc = HTTP_INTERNAL_SERVER_ERROR;
+		goto ExitFunction;
 	}
-	callback(hparser, &hparser->msg, info);
+	request = &hparser->msg;
+		#ifdef DEBUG_REDIRECT
+	getNumericHostRedirection(info->socket, host_port, sizeof host_port);
+	UpnpPrintf(UPNP_INFO,
+		MSERV,
+		__FILE__,
+		__LINE__,
+		"DEBUG TEST: Redirect host_port = %s.\n",
+		host_port);
+		#endif
+	/* chech HOST header for an IP number -- prevents DNS rebinding. */
+	if (!httpmsg_find_hdr(request, HDR_HOST, &header)) {
+		rc = UPNP_E_BAD_HTTPMSG;
+		goto ExitFunction;
+	}
+	min_size = header.length < ((sizeof host_port) - 1)
+			   ? header.length
+			   : (sizeof host_port) - 1;
+	memcpy(host_port, header.buf, min_size);
+	host_port[min_size] = 0;
+	if (host_validate_callback) {
+		rc = host_validate_callback(host_port, cookie);
+		if (rc == UPNP_E_BAD_HTTPMSG) {
+			goto ExitFunction;
+		}
+	} else if (!host_header_is_numeric(host_port, min_size)) {
+		if (!gAllowLiteralHostRedirection) {
+			rc = UPNP_E_BAD_HTTPMSG;
+			UpnpPrintf(UPNP_INFO,
+				MSERV,
+				__FILE__,
+				__LINE__,
+				"Possible DNS Rebind attack prevented.\n");
+			goto ExitFunction;
+		} else {
+			membuffer redir_buf;
+			static const char *redir_fmt =
+				"HTTP/1.1 307 Temporary Redirect\r\n"
+				"Location: http://%s\r\n\r\n";
+			char redir_str[NAME_SIZE];
+			int timeout = HTTP_DEFAULT_TIMEOUT;
 
-	return 0;
+			getNumericHostRedirection(
+				(int)info->socket, host_port, sizeof host_port);
+			membuffer_init(&redir_buf);
+			snprintf(redir_str, NAME_SIZE, redir_fmt, host_port);
+			membuffer_append_str(&redir_buf, redir_str);
+			rc = http_SendMessage(info,
+				&timeout,
+				"b",
+				redir_buf.buf,
+				redir_buf.length);
+			membuffer_destroy(&redir_buf);
+			goto ExitFunction;
+		}
+	}
+	callback(hparser, request, info);
+
+ExitFunction:
+	return rc;
 }
 
 /*!
@@ -207,8 +399,12 @@ static void handle_request(
 	struct mserv_request_t *request = (struct mserv_request_t *)args;
 	SOCKET connfd = request->connfd;
 
-	UpnpPrintf( UPNP_INFO, MSERV, __FILE__, __LINE__,
-		"miniserver %d: READING\n", connfd );
+	UpnpPrintf(UPNP_INFO,
+		MSERV,
+		__FILE__,
+		__LINE__,
+		"miniserver %d: READING\n",
+		connfd);
 	/* parser_request_init( &parser ); */ /* LEAK_FIX_MK */
 	hmsg = &parser.msg;
 	ret_code = sock_init_with_ip(
@@ -224,8 +420,12 @@ static void handle_request(
 	if (ret_code != 0) {
 		goto error_handler;
 	}
-	UpnpPrintf(UPNP_INFO, MSERV, __FILE__, __LINE__,
-		"miniserver %d: PROCESSING...\n", connfd);
+	UpnpPrintf(UPNP_INFO,
+		MSERV,
+		__FILE__,
+		__LINE__,
+		"miniserver %d: PROCESSING...\n",
+		connfd);
 	/* dispatch */
 	http_error_code = dispatch_request(&info, &parser);
 	if (http_error_code != 0) {
@@ -245,8 +445,12 @@ error_handler:
 	httpmsg_destroy(hmsg);
 	free(request);
 
-	UpnpPrintf(UPNP_INFO, MSERV, __FILE__, __LINE__,
-		"miniserver %d: COMPLETE\n", connfd);
+	UpnpPrintf(UPNP_INFO,
+		MSERV,
+		__FILE__,
+		__LINE__,
+		"miniserver %d: COMPLETE\n",
+		connfd);
 }
 
 /*!
@@ -265,29 +469,38 @@ static UPNP_INLINE void schedule_request_job(
 	memset(&job, 0, sizeof(job));
 
 	request = (struct mserv_request_t *)malloc(
-		sizeof (struct mserv_request_t));
+		sizeof(struct mserv_request_t));
 	if (request == NULL) {
-		UpnpPrintf( UPNP_INFO, MSERV, __FILE__, __LINE__,
-			"mserv %d: out of memory\n", connfd);
+		UpnpPrintf(UPNP_INFO,
+			MSERV,
+			__FILE__,
+			__LINE__,
+			"mserv %d: out of memory\n",
+			connfd);
 		sock_close(connfd);
 		return;
 	}
 
 	request->connfd = connfd;
-	memcpy(&request->foreign_sockaddr, clientAddr,
+	memcpy(&request->foreign_sockaddr,
+		clientAddr,
 		sizeof(request->foreign_sockaddr));
 	TPJobInit(&job, (start_routine)handle_request, (void *)request);
 	TPJobSetFreeFunction(&job, free_handle_request_arg);
 	TPJobSetPriority(&job, MED_PRIORITY);
 	if (ThreadPoolAdd(&gMiniServerThreadPool, &job, NULL) != 0) {
-		UpnpPrintf(UPNP_INFO, MSERV, __FILE__, __LINE__,
-			"mserv %d: cannot schedule request\n", connfd);
+		UpnpPrintf(UPNP_INFO,
+			MSERV,
+			__FILE__,
+			__LINE__,
+			"mserv %d: cannot schedule request\n",
+			connfd);
 		free(request);
 		sock_close(connfd);
 		return;
 	}
 }
-#endif
+	#endif
 
 static UPNP_INLINE void fdset_if_valid(SOCKET sock, fd_set *set)
 {
@@ -298,7 +511,7 @@ static UPNP_INLINE void fdset_if_valid(SOCKET sock, fd_set *set)
 
 static void web_server_accept(SOCKET lsock, fd_set *set)
 {
-#ifdef INTERNAL_WEB_SERVER
+	#ifdef INTERNAL_WEB_SERVER
 	SOCKET asock;
 	socklen_t clientLen;
 	struct sockaddr_storage clientAddr;
@@ -306,19 +519,22 @@ static void web_server_accept(SOCKET lsock, fd_set *set)
 
 	if (lsock != INVALID_SOCKET && FD_ISSET(lsock, set)) {
 		clientLen = sizeof(clientAddr);
-		asock = accept(lsock, (struct sockaddr *)&clientAddr,
-			&clientLen);
+		asock = accept(
+			lsock, (struct sockaddr *)&clientAddr, &clientLen);
 		if (asock == INVALID_SOCKET) {
 			strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-			UpnpPrintf(UPNP_INFO, MSERV, __FILE__, __LINE__,
+			UpnpPrintf(UPNP_INFO,
+				MSERV,
+				__FILE__,
+				__LINE__,
 				"miniserver: Error in accept(): %s\n",
 				errorBuffer);
 		} else {
-			schedule_request_job(asock,
-				(struct sockaddr *)&clientAddr);
+			schedule_request_job(
+				asock, (struct sockaddr *)&clientAddr);
 		}
 	}
-#endif /* INTERNAL_WEB_SERVER */
+	#endif /* INTERNAL_WEB_SERVER */
 }
 
 static void ssdp_read(SOCKET rsock, fd_set *set)
@@ -339,17 +555,29 @@ static int receive_from_stopSock(SOCKET ssock, fd_set *set)
 	if (FD_ISSET(ssock, set)) {
 		clientLen = sizeof(clientAddr);
 		memset((char *)&clientAddr, 0, sizeof(clientAddr));
-		byteReceived = recvfrom(ssock, requestBuf,
-			(size_t)25, 0, (struct sockaddr *)&clientAddr, &clientLen);
+		byteReceived = recvfrom(ssock,
+			requestBuf,
+			(size_t)25,
+			0,
+			(struct sockaddr *)&clientAddr,
+			&clientLen);
 		if (byteReceived > 0) {
 			requestBuf[byteReceived] = '\0';
 			inet_ntop(AF_INET,
-				&((struct sockaddr_in*)&clientAddr)->sin_addr,
-				buf_ntop, sizeof(buf_ntop));
-			UpnpPrintf( UPNP_INFO, MSERV, __FILE__, __LINE__,
+				&((struct sockaddr_in *)&clientAddr)->sin_addr,
+				buf_ntop,
+				sizeof(buf_ntop));
+			UpnpPrintf(UPNP_INFO,
+				MSERV,
+				__FILE__,
+				__LINE__,
 				"Received response: %s From host %s \n",
-				requestBuf, buf_ntop );
-			UpnpPrintf( UPNP_PACKET, MSERV, __FILE__, __LINE__,
+				requestBuf,
+				buf_ntop);
+			UpnpPrintf(UPNP_PACKET,
+				MSERV,
+				__FILE__,
+				__LINE__,
 				"Received multicast packet: \n %s\n",
 				requestBuf);
 			if (NULL != strstr(requestBuf, "ShutDown")) {
@@ -382,14 +610,15 @@ static void RunMiniServer(
 	maxMiniSock = 0;
 	maxMiniSock = max(maxMiniSock, miniSock->miniServerSock4);
 	maxMiniSock = max(maxMiniSock, miniSock->miniServerSock6);
+	maxMiniSock = max(maxMiniSock, miniSock->miniServerSock6UlaGua);
 	maxMiniSock = max(maxMiniSock, miniSock->miniServerStopSock);
 	maxMiniSock = max(maxMiniSock, miniSock->ssdpSock4);
 	maxMiniSock = max(maxMiniSock, miniSock->ssdpSock6);
 	maxMiniSock = max(maxMiniSock, miniSock->ssdpSock6UlaGua);
-#ifdef INCLUDE_CLIENT_APIS
+	#ifdef INCLUDE_CLIENT_APIS
 	maxMiniSock = max(maxMiniSock, miniSock->ssdpReqSock4);
 	maxMiniSock = max(maxMiniSock, miniSock->ssdpReqSock6);
-#endif /* INCLUDE_CLIENT_APIS */
+	#endif /* INCLUDE_CLIENT_APIS */
 	++maxMiniSock;
 
 	gMServState = MSERV_RUNNING;
@@ -401,30 +630,37 @@ static void RunMiniServer(
 		FD_SET(miniSock->miniServerStopSock, &rdSet);
 		fdset_if_valid(miniSock->miniServerSock4, &rdSet);
 		fdset_if_valid(miniSock->miniServerSock6, &rdSet);
+		fdset_if_valid(miniSock->miniServerSock6UlaGua, &rdSet);
 		fdset_if_valid(miniSock->ssdpSock4, &rdSet);
 		fdset_if_valid(miniSock->ssdpSock6, &rdSet);
 		fdset_if_valid(miniSock->ssdpSock6UlaGua, &rdSet);
-#ifdef INCLUDE_CLIENT_APIS
+	#ifdef INCLUDE_CLIENT_APIS
 		fdset_if_valid(miniSock->ssdpReqSock4, &rdSet);
 		fdset_if_valid(miniSock->ssdpReqSock6, &rdSet);
-#endif /* INCLUDE_CLIENT_APIS */
+	#endif /* INCLUDE_CLIENT_APIS */
 		/* select() */
-		ret = select((int) maxMiniSock, &rdSet, NULL, &expSet, NULL);
+		ret = select((int)maxMiniSock, &rdSet, NULL, &expSet, NULL);
 		if (ret == SOCKET_ERROR && errno == EINTR) {
 			continue;
 		}
 		if (ret == SOCKET_ERROR) {
 			strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-			UpnpPrintf(UPNP_CRITICAL, SSDP, __FILE__, __LINE__,
-				"Error in select(): %s\n", errorBuffer);
+			UpnpPrintf(UPNP_CRITICAL,
+				SSDP,
+				__FILE__,
+				__LINE__,
+				"Error in select(): %s\n",
+				errorBuffer);
 			continue;
 		} else {
 			web_server_accept(miniSock->miniServerSock4, &rdSet);
 			web_server_accept(miniSock->miniServerSock6, &rdSet);
-#ifdef INCLUDE_CLIENT_APIS
+			web_server_accept(
+				miniSock->miniServerSock6UlaGua, &rdSet);
+	#ifdef INCLUDE_CLIENT_APIS
 			ssdp_read(miniSock->ssdpReqSock4, &rdSet);
 			ssdp_read(miniSock->ssdpReqSock6, &rdSet);
-#endif /* INCLUDE_CLIENT_APIS */
+	#endif /* INCLUDE_CLIENT_APIS */
 			ssdp_read(miniSock->ssdpSock4, &rdSet);
 			ssdp_read(miniSock->ssdpSock6, &rdSet);
 			ssdp_read(miniSock->ssdpSock6UlaGua, &rdSet);
@@ -435,14 +671,15 @@ static void RunMiniServer(
 	/* Close all sockets. */
 	sock_close(miniSock->miniServerSock4);
 	sock_close(miniSock->miniServerSock6);
+	sock_close(miniSock->miniServerSock6UlaGua);
 	sock_close(miniSock->miniServerStopSock);
 	sock_close(miniSock->ssdpSock4);
 	sock_close(miniSock->ssdpSock6);
 	sock_close(miniSock->ssdpSock6UlaGua);
-#ifdef INCLUDE_CLIENT_APIS
+	#ifdef INCLUDE_CLIENT_APIS
 	sock_close(miniSock->ssdpReqSock4);
 	sock_close(miniSock->ssdpReqSock6);
-#endif /* INCLUDE_CLIENT_APIS */
+	#endif /* INCLUDE_CLIENT_APIS */
 	/* Free minisock. */
 	free(miniSock);
 	gMServState = MSERV_IDLE;
@@ -471,29 +708,273 @@ static int get_port(
 		return -1;
 	}
 	if (sockinfo.ss_family == AF_INET) {
-		*port = ntohs(((struct sockaddr_in*)&sockinfo)->sin_port);
-	} else if(sockinfo.ss_family == AF_INET6) {
-		*port = ntohs(((struct sockaddr_in6*)&sockinfo)->sin6_port);
+		*port = ntohs(((struct sockaddr_in *)&sockinfo)->sin_port);
+	} else if (sockinfo.ss_family == AF_INET6) {
+		*port = ntohs(((struct sockaddr_in6 *)&sockinfo)->sin6_port);
 	}
-	UpnpPrintf(UPNP_INFO, MSERV, __FILE__, __LINE__,
-		"sockfd = %d, .... port = %d\n", sockfd, (int)*port);
+	UpnpPrintf(UPNP_INFO,
+		MSERV,
+		__FILE__,
+		__LINE__,
+		"sockfd = %d, .... port = %d\n",
+		sockfd,
+		(int)*port);
 
 	return 0;
 }
 
-#ifdef INTERNAL_WEB_SERVER
+	#ifdef INTERNAL_WEB_SERVER
+static int init_socket_suff(
+	struct s_SocketStuff *s, const char *text_addr, int ip_version)
+{
+	char errorBuffer[ERROR_BUFFER_LEN];
+	int sockError;
+	sa_family_t domain;
+	void *addr;
+	int reuseaddr_on = MINISERVER_REUSEADDR;
+
+	memset(s, 0, sizeof *s);
+	s->fd = INVALID_SOCKET;
+	s->ip_version = ip_version;
+	s->text_addr = text_addr;
+	s->serverAddr = (struct sockaddr *)&s->ss;
+	switch (ip_version) {
+	case 4:
+		domain = AF_INET;
+		s->serverAddr4->sin_family = domain;
+		s->address_len = sizeof *s->serverAddr4;
+		addr = &s->serverAddr4->sin_addr;
+		break;
+	case 6:
+		if (!ENABLE_IPV6) {
+			goto ok;
+		}
+		domain = AF_INET6;
+		s->serverAddr6->sin6_family = domain;
+		s->address_len = sizeof *s->serverAddr6;
+		addr = &s->serverAddr6->sin6_addr;
+		break;
+	default:
+		UpnpPrintf(UPNP_INFO,
+			MSERV,
+			__FILE__,
+			__LINE__,
+			"init_socket_suff(): Invalid IP version: %d.\n",
+			ip_version);
+		goto error;
+		break;
+	}
+
+	if (inet_pton(domain, text_addr, addr) <= 0)
+		goto error;
+
+	s->fd = socket(domain, SOCK_STREAM, 0);
+	if (s->fd == INVALID_SOCKET) {
+		strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
+		UpnpPrintf(UPNP_INFO,
+			MSERV,
+			__FILE__,
+			__LINE__,
+			"init_socket_suff(): IPv%c socket not available: "
+			"%s\n",
+			ip_version,
+			errorBuffer);
+		goto error;
+	} else if (ip_version == 6) {
+		int onOff = 1;
+
+		sockError = setsockopt(s->fd,
+			IPPROTO_IPV6,
+			IPV6_V6ONLY,
+			(char *)&onOff,
+			sizeof(onOff));
+		if (sockError == SOCKET_ERROR) {
+			strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
+			UpnpPrintf(UPNP_INFO,
+				MSERV,
+				__FILE__,
+				__LINE__,
+				"init_socket_suff(): unable to set IPv6 "
+				"socket protocol: %s\n",
+				errorBuffer);
+			goto error;
+		}
+	}
+	/* Getting away with implementation of re-using address:port and
+	 * instead choosing to increment port numbers.
+	 * Keeping the re-use address code as an optional behaviour that
+	 * can be turned on if necessary.
+	 * TURN ON the reuseaddr_on option to use the option. */
+	if (MINISERVER_REUSEADDR) {
+		sockError = setsockopt(s->fd,
+			SOL_SOCKET,
+			SO_REUSEADDR,
+			(const char *)&reuseaddr_on,
+			sizeof(int));
+		if (sockError == SOCKET_ERROR) {
+			strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
+			UpnpPrintf(UPNP_INFO,
+				MSERV,
+				__FILE__,
+				__LINE__,
+				"init_socket_suff(): unable to set "
+				"SO_REUSEADDR: %s\n",
+				errorBuffer);
+			goto error;
+		}
+	}
+ok:
+	return 0;
+
+error:
+	if (s->fd != INVALID_SOCKET) {
+		sock_close(s->fd);
+	}
+	s->fd = INVALID_SOCKET;
+
+	return 1;
+}
+
+/*
+ * s->port will be one more than the used port in the end. This is important,
+ * in case this function is called again.
+ */
+static int do_bind(struct s_SocketStuff *s)
+{
+	int ret_val = UPNP_E_SUCCESS;
+	int bind_error;
+	int errCode = 0;
+	char errorBuffer[ERROR_BUFFER_LEN];
+	uint16_t original_listen_port = s->try_port;
+
+	do {
+		switch (s->ip_version) {
+		case 4:
+			s->serverAddr4->sin_port = htons(s->try_port++);
+			break;
+		case 6:
+			s->serverAddr6->sin6_port = htons(s->try_port++);
+			break;
+		}
+		bind_error = bind(s->fd, s->serverAddr, s->address_len);
+		if (bind_error == SOCKET_ERROR) {
+		#ifdef _WIN32
+			errCode = WSAGetLastError();
+		#else
+			errCode = errno;
+		#endif
+			if (errno == EADDRINUSE) {
+				errCode = 1;
+			}
+		} else {
+			errCode = 0;
+		}
+	} while (errCode != 0 && s->try_port >= original_listen_port);
+	if (bind_error == SOCKET_ERROR) {
+		strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
+		UpnpPrintf(UPNP_INFO,
+			MSERV,
+			__FILE__,
+			__LINE__,
+			"get_miniserver_sockets: "
+			"Error in IPv%d bind(): %s\n",
+			s->ip_version,
+			errorBuffer);
+		/* Bind failied. */
+		ret_val = UPNP_E_SOCKET_BIND;
+		goto error;
+	}
+
+	return UPNP_E_SUCCESS;
+
+error:
+	return ret_val;
+}
+
+static int do_listen(struct s_SocketStuff *s)
+{
+	int ret_val;
+	int listen_error;
+	int port_error;
+	char errorBuffer[ERROR_BUFFER_LEN];
+
+	listen_error = listen(s->fd, SOMAXCONN);
+	if (listen_error == SOCKET_ERROR) {
+		strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
+		UpnpPrintf(UPNP_INFO,
+			MSERV,
+			__FILE__,
+			__LINE__,
+			"do_listen(): Error in IPv%d listen(): %s\n",
+			s->ip_version,
+			errorBuffer);
+		ret_val = UPNP_E_LISTEN;
+		goto error;
+	}
+	port_error = get_port(s->fd, &s->actual_port);
+	if (port_error < 0) {
+		UpnpPrintf(UPNP_INFO,
+			MSERV,
+			__FILE__,
+			__LINE__,
+			"do_listen(): Error in get_port().\n");
+		ret_val = UPNP_E_INTERNAL_ERROR;
+		goto error;
+	}
+
+	return UPNP_E_SUCCESS;
+
+error:
+	return ret_val;
+}
+
+static int do_reinit(struct s_SocketStuff *s)
+{
+	sock_close(s->fd);
+
+	return init_socket_suff(s, s->text_addr, s->ip_version);
+}
+
+static int do_bind_listen(struct s_SocketStuff *s)
+{
+	int ret_val;
+	int ok = 0;
+	int original_port = s->try_port;
+
+	while (!ok) {
+		ret_val = do_bind(s);
+		if (ret_val) {
+			goto error;
+		}
+		ret_val = do_listen(s);
+		if (ret_val) {
+			if (errno == EADDRINUSE) {
+				do_reinit(s);
+				continue;
+			}
+			goto error;
+		}
+		ok = s->try_port >= original_port;
+	}
+
+	return 0;
+
+error:
+	return ret_val;
+}
+
 /*!
  * \brief Creates a STREAM socket, binds to INADDR_ANY and listens for
  * incoming connecttions. Returns the actual port which the sockets
- * sub-system returned. 
+ * sub-system returned.
  *
- * Also creates a DGRAM socket, binds to the loop back address and 
+ * Also creates a DGRAM socket, binds to the loop back address and
  * returns the port allocated by the socket sub-system.
  *
  * \return
  *	\li UPNP_E_OUTOF_SOCKET: Failed to create a socket.
  *	\li UPNP_E_SOCKET_BIND: Bind() failed.
- *	\li UPNP_E_LISTEN: Listen() failed.	
+ *	\li UPNP_E_LISTEN: Listen() failed.
  *	\li UPNP_E_INTERNAL_ERROR: Port returned by the socket layer is < 0.
  *	\li UPNP_E_SUCCESS: Success.
  */
@@ -502,266 +983,117 @@ static int get_miniserver_sockets(
 	MiniServerSockArray *out,
 	/*! [in] port on which the server is listening for incoming IPv4
 	 * connections. */
-	uint16_t listen_port4
-#ifdef UPNP_ENABLE_IPV6
-	,
+	uint16_t listen_port4,
 	/*! [in] port on which the server is listening for incoming IPv6
-	 * connections. */
-	uint16_t listen_port6
-#endif
-	)
+	 * ULA connections. */
+	uint16_t listen_port6,
+	/*! [in] port on which the server is listening for incoming
+	 * IPv6 ULA or GUA connections. */
+	uint16_t listen_port6UlaGua)
 {
-	char errorBuffer[ERROR_BUFFER_LEN];
-	struct sockaddr_storage __ss_v4;
-	struct sockaddr_in* serverAddr4 = (struct sockaddr_in*)&__ss_v4;
-	SOCKET listenfd4;
-	uint16_t actual_port4 = 0u;
-#ifdef UPNP_ENABLE_IPV6
-	struct sockaddr_storage __ss_v6;
-	struct sockaddr_in6* serverAddr6 = (struct sockaddr_in6*)&__ss_v6;
-	SOCKET listenfd6;
-	uint16_t actual_port6 = 0u;
-#endif
-	int ret_code;
-#ifdef UPNP_MINISERVER_REUSEADDR
-	int reuseaddr_on = 1;
-#else
-	int reuseaddr_on = 0;
-#endif
+	int ret_val;
+	int err_init_4;
+	int err_init_6;
+	int err_init_6UlaGua;
 
-	int sockError = UPNP_E_SUCCESS;
-	int errCode = 0;
+	struct s_SocketStuff ss4;
+	struct s_SocketStuff ss6;
+	struct s_SocketStuff ss6UlaGua;
 
 	/* Create listen socket for IPv4/IPv6. An error here may indicate
 	 * that we don't have an IPv4/IPv6 stack. */
-	listenfd4 = socket(AF_INET, SOCK_STREAM, 0);
-	if (listenfd4 == INVALID_SOCKET) {
-		strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-		UpnpPrintf(UPNP_INFO, MSERV, __FILE__, __LINE__,
-			"get_miniserver_sockets: IPv4 socket not available: %s\n",
-			errorBuffer);
-	}
-#ifdef UPNP_ENABLE_IPV6
-	listenfd6 = socket(AF_INET6, SOCK_STREAM, 0);
-	if (listenfd6 == INVALID_SOCKET) {
-		strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-		UpnpPrintf(UPNP_INFO, MSERV, __FILE__, __LINE__,
-			"get_miniserver_sockets: IPv6 socket not available: %s\n",
-			errorBuffer);
-	} else {
-		int onOff = 1;
-		sockError = setsockopt(listenfd6, IPPROTO_IPV6, IPV6_V6ONLY,
-							   (char *)&onOff, sizeof(onOff));
-		if (sockError == SOCKET_ERROR) {
-			strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-			UpnpPrintf(UPNP_INFO, MSERV, __FILE__, __LINE__,
-				"get_miniserver_sockets: unable to set IPv6 socket protocol: %s\n",
-				errorBuffer);
-			sock_close(listenfd6);
-			listenfd6 = INVALID_SOCKET;
-		}
-	}
-#endif
-	if (listenfd4 == INVALID_SOCKET
-#ifdef UPNP_ENABLE_IPV6
-	    && listenfd6 == INVALID_SOCKET
-#endif
-	) {
-		UpnpPrintf(UPNP_CRITICAL, MSERV, __FILE__, __LINE__,
+	err_init_4 = init_socket_suff(&ss4, gIF_IPV4, 4);
+	err_init_6 = init_socket_suff(&ss6, gIF_IPV6, 6);
+	err_init_6UlaGua = init_socket_suff(&ss6UlaGua, gIF_IPV6_ULA_GUA, 6);
+	ss6.serverAddr6->sin6_scope_id = gIF_INDEX;
+	/* Check what happened. */
+	if (err_init_4 && (err_init_6 || err_init_6UlaGua)) {
+		UpnpPrintf(UPNP_CRITICAL,
+			MSERV,
+			__FILE__,
+			__LINE__,
 			"get_miniserver_sockets: no protocols available\n");
-		return UPNP_E_OUTOF_SOCKET;
+		ret_val = UPNP_E_OUTOF_SOCKET;
+		goto error;
 	}
 	/* As per the IANA specifications for the use of ports by applications
 	 * override the listen port passed in with the first available. */
-	if (listen_port4 < APPLICATION_LISTENING_PORT) {
+	if (listen_port4 == 0) {
 		listen_port4 = (uint16_t)APPLICATION_LISTENING_PORT;
 	}
-#ifdef UPNP_ENABLE_IPV6
-	if (listen_port6 < APPLICATION_LISTENING_PORT) {
+	if (listen_port6 == 0) {
 		listen_port6 = (uint16_t)APPLICATION_LISTENING_PORT;
 	}
-#endif
-	memset(&__ss_v4, 0, sizeof (__ss_v4));
-	serverAddr4->sin_family = (sa_family_t)AF_INET;
-	inet_pton(AF_INET, gIF_IPV4, &serverAddr4->sin_addr);
-#ifdef UPNP_ENABLE_IPV6
-	memset(&__ss_v6, 0, sizeof (__ss_v6));
-	serverAddr6->sin6_family = (sa_family_t)AF_INET6;
-	inet_pton(AF_INET6, gIF_IPV6, &serverAddr6->sin6_addr);
-	serverAddr6->sin6_scope_id = gIF_INDEX;
-#endif
-	/* Getting away with implementation of re-using address:port and
-	 * instead choosing to increment port numbers.
-	 * Keeping the re-use address code as an optional behaviour that
-	 * can be turned on if necessary.
-	 * TURN ON the reuseaddr_on option to use the option. */
-	if (reuseaddr_on) {
+	if (listen_port6UlaGua < APPLICATION_LISTENING_PORT) {
+		/* Increment the port to make it harder to fail at first try */
+		listen_port6UlaGua = listen_port6 + 1;
+	}
+	ss4.try_port = listen_port4;
+	ss6.try_port = listen_port6;
+	ss6UlaGua.try_port = listen_port6UlaGua;
+	if (MINISERVER_REUSEADDR) {
 		/* THIS IS ALLOWS US TO BIND AGAIN IMMEDIATELY
 		 * AFTER OUR SERVER HAS BEEN CLOSED
 		 * THIS MAY CAUSE TCP TO BECOME LESS RELIABLE
 		 * HOWEVER IT HAS BEEN SUGESTED FOR TCP SERVERS. */
-		UpnpPrintf(UPNP_INFO, MSERV, __FILE__, __LINE__,
+		UpnpPrintf(UPNP_INFO,
+			MSERV,
+			__FILE__,
+			__LINE__,
 			"get_miniserver_sockets: resuseaddr is set.\n");
-		if (listenfd4 != INVALID_SOCKET) {
-			sockError = setsockopt(listenfd4, SOL_SOCKET,
-				SO_REUSEADDR,
-				(const char *)&reuseaddr_on, sizeof (int));
-			if (sockError == SOCKET_ERROR) {
-				sock_close(listenfd4);
-#ifdef UPNP_ENABLE_IPV6
-				sock_close(listenfd6);
-#endif
-				return UPNP_E_SOCKET_BIND;
-			}
-		}
-#ifdef UPNP_ENABLE_IPV6
-		if (listenfd6 != INVALID_SOCKET) {
-			sockError = setsockopt(listenfd6, SOL_SOCKET,
-				SO_REUSEADDR,
-			(const char *)&reuseaddr_on, sizeof (int));
-			if (sockError == SOCKET_ERROR) {
-				sock_close(listenfd4);
-				sock_close(listenfd6);
-				return UPNP_E_SOCKET_BIND;
-			}
-		}
-#endif  /* IPv6 */
 	}
-	
-	if (listenfd4 != INVALID_SOCKET) {
-		uint16_t orig_listen_port4 = listen_port4;
-		do {
-			serverAddr4->sin_port = htons(listen_port4++);
-			sockError = bind(listenfd4,
-					 (struct sockaddr *)serverAddr4,
-					 sizeof(*serverAddr4));
-			if (sockError == SOCKET_ERROR) {
-#ifdef _WIN32
-				errCode = WSAGetLastError();
-#else
-				errCode = errno;
-#endif
-				if (errno == EADDRINUSE) {
-					errCode = 1;
-				}
-			} else {
-				errCode = 0;
-			}
-		} while (errCode != 0 &&
-			 listen_port4 >= orig_listen_port4);
-		if (sockError == SOCKET_ERROR) {
-			strerror_r(errno, errorBuffer,
-				   ERROR_BUFFER_LEN);
-			UpnpPrintf(UPNP_INFO, MSERV,
-				   __FILE__, __LINE__,
-				   "get_miniserver_sockets: "
-				   "Error in IPv4 bind(): %s\n",
-				   errorBuffer);
-			sock_close(listenfd4);
-#ifdef UPNP_ENABLE_IPV6
-			sock_close(listenfd6);
-#endif
-			/* Bind failied. */
-			return UPNP_E_SOCKET_BIND;
+	if (ss4.fd != INVALID_SOCKET) {
+		ret_val = do_bind_listen(&ss4);
+		if (ret_val) {
+			goto error;
 		}
 	}
-#ifdef UPNP_ENABLE_IPV6
-	if (listenfd6 != INVALID_SOCKET) {
-		uint16_t orig_listen_port6 = listen_port6;
-		do {
-			serverAddr6->sin6_port = htons(listen_port6++);
-			sockError = bind(listenfd6,
-					 (struct sockaddr *)serverAddr6,
-					 sizeof(*serverAddr6));
-			if (sockError == SOCKET_ERROR) {
-#ifdef _WIN32
-				errCode = WSAGetLastError();
-#else
-				errCode = errno; 
-#endif
-				if (errno == EADDRINUSE) {
-					errCode = 1;
-				}
-			} else {
-				errCode = 0;
-			}
-		} while (errCode != 0 &&
-			 listen_port6 >= orig_listen_port6);
-		if (sockError == SOCKET_ERROR) {
-			strerror_r(errno, errorBuffer,
-				   ERROR_BUFFER_LEN);
-			UpnpPrintf(UPNP_INFO, MSERV,
-				   __FILE__, __LINE__,
-				   "get_miniserver_sockets: "
-				   "Error in IPv6 bind(): %s\n",
-				   errorBuffer);
-			sock_close(listenfd4);
-			sock_close(listenfd6);
-			/* Bind failied. */
-			return UPNP_E_SOCKET_BIND;
+	if (ss6.fd != INVALID_SOCKET) {
+		ret_val = do_bind_listen(&ss6);
+		if (ret_val) {
+			goto error;
 		}
 	}
-#endif
-
-	UpnpPrintf(UPNP_INFO, MSERV, __FILE__, __LINE__,
+	if (ss6UlaGua.fd != INVALID_SOCKET) {
+		ret_val = do_bind_listen(&ss6UlaGua);
+		if (ret_val) {
+			goto error;
+		}
+	}
+	UpnpPrintf(UPNP_INFO,
+		MSERV,
+		__FILE__,
+		__LINE__,
 		"get_miniserver_sockets: bind successful\n");
-	if (listenfd4 != INVALID_SOCKET) {
-		ret_code = listen(listenfd4, SOMAXCONN);
-		if (ret_code == SOCKET_ERROR) {
-			strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-			UpnpPrintf(UPNP_INFO, MSERV, __FILE__, __LINE__,
-				"mserv start: Error in IPv4 listen(): %s\n",
-				errorBuffer);
-			sock_close(listenfd4);
-#ifdef UPNP_ENABLE_IPV6
-			sock_close(listenfd6);
-#endif
-			return UPNP_E_LISTEN;
-		}
-		ret_code = get_port(listenfd4, &actual_port4);
-		if (ret_code < 0) {
-			sock_close(listenfd4);
-#ifdef UPNP_ENABLE_IPV6
-			sock_close(listenfd6);
-#endif
-			return UPNP_E_INTERNAL_ERROR;
-		}
-		out->miniServerPort4 = actual_port4;
-	}
-#ifdef UPNP_ENABLE_IPV6
-	if (listenfd6 != INVALID_SOCKET) {
-		ret_code = listen(listenfd6, SOMAXCONN);
-		if (ret_code == SOCKET_ERROR) {
-			strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-			UpnpPrintf(UPNP_INFO, MSERV, __FILE__, __LINE__,
-				"mserv start: Error in IPv6 listen(): %s\n",
-				errorBuffer);
-			sock_close(listenfd4);
-			sock_close(listenfd6);
-			return UPNP_E_LISTEN;
-		}
-		ret_code = get_port(listenfd6, &actual_port6);
-		if (ret_code < 0) {
-			sock_close(listenfd4);
-			sock_close(listenfd6);
-			return UPNP_E_INTERNAL_ERROR;
-		}
-		out->miniServerPort6 = actual_port6;
-	}
-#endif
-	out->miniServerSock4 = listenfd4;
-#ifdef UPNP_ENABLE_IPV6
-	out->miniServerSock6 = listenfd6;
-#endif
+	out->miniServerPort4 = ss4.actual_port;
+	out->miniServerPort6 = ss6.actual_port;
+	out->miniServerPort6UlaGua = ss6UlaGua.actual_port;
+	out->miniServerSock4 = ss4.fd;
+	out->miniServerSock6 = ss6.fd;
+	out->miniServerSock6UlaGua = ss6UlaGua.fd;
+
 	return UPNP_E_SUCCESS;
+
+error:
+	if (ss4.fd != INVALID_SOCKET) {
+		sock_close(ss4.fd);
+	}
+	if (ss6.fd != INVALID_SOCKET) {
+		sock_close(ss6.fd);
+	}
+	if (ss6UlaGua.fd != INVALID_SOCKET) {
+		sock_close(ss6UlaGua.fd);
+	}
+
+	return ret_val;
 }
-#endif /* INTERNAL_WEB_SERVER */
+	#endif /* INTERNAL_WEB_SERVER */
 
 /*!
- * \brief Creates the miniserver STOP socket. This socket is created and 
+ * \brief Creates the miniserver STOP socket. This socket is created and
  *  listened on to know when it is time to stop the Miniserver.
  *
- * \return 
+ * \return
  * \li \c UPNP_E_OUTOF_SOCKET: Failed to create a socket.
  * \li \c UPNP_E_SOCKET_BIND: Bind() failed.
  * \li \c UPNP_E_INTERNAL_ERROR: Port returned by the socket layer is < 0.
@@ -779,19 +1111,30 @@ static int get_miniserver_stopsock(
 	miniServerStopSock = socket(AF_INET, SOCK_DGRAM, 0);
 	if (miniServerStopSock == INVALID_SOCKET) {
 		strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-		UpnpPrintf(UPNP_CRITICAL, MSERV, __FILE__, __LINE__,
-			"Error in socket(): %s\n", errorBuffer);
+		UpnpPrintf(UPNP_CRITICAL,
+			MSERV,
+			__FILE__,
+			__LINE__,
+			"Error in socket(): %s\n",
+			errorBuffer);
 		return UPNP_E_OUTOF_SOCKET;
 	}
 	/* Bind to local socket. */
-	memset(&stop_sockaddr, 0, sizeof (stop_sockaddr));
+	memset(&stop_sockaddr, 0, sizeof(stop_sockaddr));
 	stop_sockaddr.sin_family = (sa_family_t)AF_INET;
+	#ifdef _WIN32
+	inet_pton(AF_INET, "127.0.0.1", &stop_sockaddr.sin_addr);
+	#else
 	stop_sockaddr.sin_addr.s_addr = inet_addr("127.0.0.1");
-	ret = bind(miniServerStopSock, (struct sockaddr *)&stop_sockaddr,
+	#endif
+	ret = bind(miniServerStopSock,
+		(struct sockaddr *)&stop_sockaddr,
 		sizeof(stop_sockaddr));
 	if (ret == SOCKET_ERROR) {
 		UpnpPrintf(UPNP_CRITICAL,
-		MSERV, __FILE__, __LINE__,
+			MSERV,
+			__FILE__,
+			__LINE__,
 			"Error in binding localhost!!!\n");
 		sock_close(miniServerStopSock);
 		return UPNP_E_SOCKET_BIND;
@@ -811,6 +1154,7 @@ static UPNP_INLINE void InitMiniServerSockArray(MiniServerSockArray *miniSocket)
 {
 	miniSocket->miniServerSock4 = INVALID_SOCKET;
 	miniSocket->miniServerSock6 = INVALID_SOCKET;
+	miniSocket->miniServerSock6UlaGua = INVALID_SOCKET;
 	miniSocket->miniServerStopSock = INVALID_SOCKET;
 	miniSocket->ssdpSock4 = INVALID_SOCKET;
 	miniSocket->ssdpSock6 = INVALID_SOCKET;
@@ -818,19 +1162,23 @@ static UPNP_INLINE void InitMiniServerSockArray(MiniServerSockArray *miniSocket)
 	miniSocket->stopPort = 0u;
 	miniSocket->miniServerPort4 = 0u;
 	miniSocket->miniServerPort6 = 0u;
-#ifdef INCLUDE_CLIENT_APIS
+	miniSocket->miniServerPort6UlaGua = 0u;
+	#ifdef INCLUDE_CLIENT_APIS
 	miniSocket->ssdpReqSock4 = INVALID_SOCKET;
 	miniSocket->ssdpReqSock6 = INVALID_SOCKET;
-#endif /* INCLUDE_CLIENT_APIS */
+	#endif /* INCLUDE_CLIENT_APIS */
 }
 
 int StartMiniServer(
 	/*! [in,out] Port on which the server listens for incoming IPv4
 	 * connections. */
-	uint16_t *listen_port4, 
+	uint16_t *listen_port4,
 	/*! [in,out] Port on which the server listens for incoming IPv6
-	 * connections. */
-	uint16_t *listen_port6)
+	 * LLA connections. */
+	uint16_t *listen_port6,
+	/*! [in,out] Port on which the server listens for incoming
+	 * IPv6 ULA or GUA connections. */
+	uint16_t *listen_port6UlaGua)
 {
 	int ret_code;
 	int count;
@@ -847,30 +1195,26 @@ int StartMiniServer(
 		/* miniserver running. */
 		return UPNP_E_INTERNAL_ERROR;
 	}
-	miniSocket = (MiniServerSockArray *)malloc(
-		sizeof (MiniServerSockArray));
+	miniSocket = (MiniServerSockArray *)malloc(sizeof(MiniServerSockArray));
 	if (!miniSocket) {
 		return UPNP_E_OUTOF_MEMORY;
 	}
 	InitMiniServerSockArray(miniSocket);
-#ifdef INTERNAL_WEB_SERVER
+	#ifdef INTERNAL_WEB_SERVER
 	/* V4 and V6 http listeners. */
 	ret_code = get_miniserver_sockets(
-		miniSocket, *listen_port4
-#ifdef UPNP_ENABLE_IPV6
-		, *listen_port6
-#endif
-		);
+		miniSocket, *listen_port4, *listen_port6, *listen_port6UlaGua);
 	if (ret_code != UPNP_E_SUCCESS) {
 		free(miniSocket);
 		return ret_code;
 	}
-#endif
+	#endif
 	/* Stop socket (To end miniserver processing). */
 	ret_code = get_miniserver_stopsock(miniSocket);
 	if (ret_code != UPNP_E_SUCCESS) {
 		sock_close(miniSocket->miniServerSock4);
 		sock_close(miniSocket->miniServerSock6);
+		sock_close(miniSocket->miniServerSock6UlaGua);
 		free(miniSocket);
 		return ret_code;
 	}
@@ -879,6 +1223,7 @@ int StartMiniServer(
 	if (ret_code != UPNP_E_SUCCESS) {
 		sock_close(miniSocket->miniServerSock4);
 		sock_close(miniSocket->miniServerSock6);
+		sock_close(miniSocket->miniServerSock6UlaGua);
 		sock_close(miniSocket->miniServerStopSock);
 		free(miniSocket);
 		return ret_code;
@@ -890,20 +1235,22 @@ int StartMiniServer(
 	if (ret_code < 0) {
 		sock_close(miniSocket->miniServerSock4);
 		sock_close(miniSocket->miniServerSock6);
+		sock_close(miniSocket->miniServerSock6UlaGua);
 		sock_close(miniSocket->miniServerStopSock);
 		sock_close(miniSocket->ssdpSock4);
 		sock_close(miniSocket->ssdpSock6);
 		sock_close(miniSocket->ssdpSock6UlaGua);
-#ifdef INCLUDE_CLIENT_APIS
+	#ifdef INCLUDE_CLIENT_APIS
 		sock_close(miniSocket->ssdpReqSock4);
 		sock_close(miniSocket->ssdpReqSock6);
-#endif /* INCLUDE_CLIENT_APIS */
+	#endif /* INCLUDE_CLIENT_APIS */
 		free(miniSocket);
 		return UPNP_E_OUTOF_MEMORY;
 	}
 	/* Wait for miniserver to start. */
 	count = 0;
-	while (gMServState != (MiniServerState)MSERV_RUNNING && count < max_count) {
+	while (gMServState != (MiniServerState)MSERV_RUNNING &&
+		count < max_count) {
 		/* 0.05s */
 		imillisleep(50);
 		count++;
@@ -912,20 +1259,22 @@ int StartMiniServer(
 		/* Took it too long to start that thread. */
 		sock_close(miniSocket->miniServerSock4);
 		sock_close(miniSocket->miniServerSock6);
+		sock_close(miniSocket->miniServerSock6UlaGua);
 		sock_close(miniSocket->miniServerStopSock);
 		sock_close(miniSocket->ssdpSock4);
 		sock_close(miniSocket->ssdpSock6);
 		sock_close(miniSocket->ssdpSock6UlaGua);
-#ifdef INCLUDE_CLIENT_APIS
+	#ifdef INCLUDE_CLIENT_APIS
 		sock_close(miniSocket->ssdpReqSock4);
 		sock_close(miniSocket->ssdpReqSock6);
-#endif /* INCLUDE_CLIENT_APIS */
+	#endif /* INCLUDE_CLIENT_APIS */
 		return UPNP_E_INTERNAL_ERROR;
 	}
-#ifdef INTERNAL_WEB_SERVER
+	#ifdef INTERNAL_WEB_SERVER
 	*listen_port4 = miniSocket->miniServerPort4;
 	*listen_port6 = miniSocket->miniServerPort6;
-#endif
+	*listen_port6UlaGua = miniSocket->miniServerPort6UlaGua;
+	#endif
 
 	return UPNP_E_SUCCESS;
 }
@@ -933,13 +1282,13 @@ int StartMiniServer(
 int StopMiniServer()
 {
 	char errorBuffer[ERROR_BUFFER_LEN];
-	socklen_t socklen = sizeof (struct sockaddr_in);
+	socklen_t socklen = sizeof(struct sockaddr_in);
 	SOCKET sock;
 	struct sockaddr_in ssdpAddr;
 	char buf[256] = "ShutDown";
 	size_t bufLen = strlen(buf);
 
-	switch(gMServState) {
+	switch (gMServState) {
 	case MSERV_RUNNING:
 		gMServState = MSERV_STOPPING;
 		break;
@@ -949,17 +1298,28 @@ int StopMiniServer()
 	sock = socket(AF_INET, SOCK_DGRAM, 0);
 	if (sock == INVALID_SOCKET) {
 		strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-		UpnpPrintf(UPNP_INFO, SSDP, __FILE__, __LINE__,
+		UpnpPrintf(UPNP_INFO,
+			SSDP,
+			__FILE__,
+			__LINE__,
 			"SSDP_SERVER: StopSSDPServer: Error in socket() %s\n",
 			errorBuffer);
 		return 0;
 	}
-	while(gMServState != (MiniServerState)MSERV_IDLE) {
+	while (gMServState != (MiniServerState)MSERV_IDLE) {
 		ssdpAddr.sin_family = (sa_family_t)AF_INET;
+	#ifdef _WIN32
+		inet_pton(AF_INET, "127.0.0.1", &ssdpAddr.sin_addr);
+	#else
 		ssdpAddr.sin_addr.s_addr = inet_addr("127.0.0.1");
+	#endif
 		ssdpAddr.sin_port = htons(miniStopSockPort);
-		sendto(sock, buf, bufLen, 0,
-			(struct sockaddr *)&ssdpAddr, socklen);
+		sendto(sock,
+			buf,
+			bufLen,
+			0,
+			(struct sockaddr *)&ssdpAddr,
+			socklen);
 		imillisleep(1);
 		if (gMServState == (MiniServerState)MSERV_IDLE) {
 			break;
